import logging
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple

from brownie.convert.datatypes import EthAddress
from hexbytes import HexBytes
from y.classes.common import ContractBase
from y.contracts import build_name
from y.decorators import log
from y.typing import Block
from y.utils.events import decode_logs, get_logs_asap
from y.utils.multicall import fetch_multicall

logger = logging.getLogger(__name__)


class BalancerV2Vault(ContractBase):
    def __init__(self, address: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
        if not self._is_cached:
            # we need the contract cached so we can decode logs correctly
            self.contract
    
    @log(logger)
    def get_pool_tokens(self, pool_id: int, block: Optional[Block] = None):
        return self.contract.getPoolTokens(pool_id, block_identifier = block)

    @log(logger)
    @lru_cache(maxsize=10)
    def list_pools(self, block: Optional[Block] = None) -> Dict[HexBytes,EthAddress]:
        topics = ['0x3c13bc30b8e878c53fd2a36b679409c073afd75950be43d8858768e956fbc20e']
        try:
            events = decode_logs(get_logs_asap(self.address, topics, to_block=block))
        except TypeError as e:
            if "Start must be less than or equal to stop" in str(e):
                return {}
            raise
        return {event['poolId'].hex():event['poolAddress'] for event in events}
    
    @lru_cache(maxsize=10)
    def get_pool_info(self, poolids: Tuple[HexBytes,...], block: Optional[Block] = None) -> List[Tuple]:
        return fetch_multicall(*[[self.contract,'getPoolTokens',poolId] for poolId in poolids], block=block)

    @log(logger)
    def deepest_pool_for(self, token_address: EthAddress, block: Optional[Block] = None) -> Tuple[Optional[EthAddress],int]:
        pools = self.list_pools(block=block)
        poolids = (poolid for poolid, pool in pools.items() if _is_standard_pool(pool))
        pools_info = self.get_pool_info(poolids, block=block)
        pools_info = {self.list_pools(block=block)[poolid]: info for poolid, info in zip(poolids, pools_info) if str(info) != "((), (), 0)"}
        
        deepest_pool = {'pool': None, 'balance': 0}
        for pool, info in pools_info.items():
            num_tokens = len(info[0])
            pool_balances = {info[0][i]: info[1][i] for i in range(num_tokens)}
            pool_balance = [balance for token, balance in pool_balances.items() if token == token_address]
            if len(pool_balance) == 0:
                continue
            assert len(pool_balance) == 1
            pool_balance = pool_balance[0]
            if pool_balance > deepest_pool['balance']:
                deepest_pool = {'pool': pool, 'balance': pool_balance}

        return deepest_pool['pool'], deepest_pool['balance']


@log(logger)
@lru_cache(maxsize=None)
def _is_standard_pool(pool: EthAddress) -> bool:
    '''
    Returns `False` if `build_name(pool) in ['ConvergentCurvePool','MetaStablePool']`, else `True`
    '''
    
    # With `return_None_on_failure=True`, if `build_name(pool)` fails,
    # we can't know for sure that its a standard pool, but... it probably is.
    return build_name(pool, return_None_on_failure=True) not in ['ConvergentCurvePool','MetaStablePool']



import logging
from typing import Dict, Optional

from brownie import chain
from cachetools.func import ttl_cache
from y import convert
from y.datatypes import UsdPrice
from y.decorators import log
from y.exceptions import contract_not_verified
from y.networks import Network
from y.typing import Address, AnyAddressType, Block
from y.uniswap.protocols import UNISWAPS
from y.uniswap.v1 import UniswapV1
from y.uniswap.v2.pool import NotAUniswapV2Pool, UniswapPoolV2
from y.uniswap.v2.router import UniswapRouterV2
from y.utils.logging import gh_issue_request
from y.utils.multicall import multicall_same_func_no_input

logger = logging.getLogger(__name__)

# NOTE: If this is failing to pull a price for a token you need, it's likely because that token requires a special swap path.
#       Please add a viable swap path to ..protocols to fetch price data successfully.

class Uniswap:
    def __init__(self) -> None:
        self.routers = {}
        for name in UNISWAPS:
            try: self.routers[name] = UniswapRouterV2(UNISWAPS[name]['router'])
            except ValueError as e: # TODO do this better
                if not contract_not_verified(e):
                    raise
        self.factories = [UNISWAPS[name]['factory'] for name in UNISWAPS]
        self.v1 = UniswapV1()

    @log(logger)
    def is_uniswap_pool(self, token_address: AnyAddressType) -> bool:
        token_address = convert.to_address(token_address)
        try:
            pool = UniswapPoolV2(token_address)
            is_pool = all(pool.get_pool_details())
            if is_pool and pool.factory not in self.factories:
                gh_issue_request(f'UniClone Factory {pool.factory} is unknown to ypricemagic.', logger)
                self.factories.append(pool.factory)
            return is_pool

        except NotAUniswapV2Pool: return False

    @log(logger)
    def get_price_v1(self, token_address: Address, block: Optional[Block] = None) -> UsdPrice:
        return self.v1.get_price(token_address, block)
    
    @log(logger)
    @ttl_cache(ttl=600)
    def lp_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        """ Get Uniswap/Sushiswap LP token price. """
        return UniswapPoolV2(token_address).get_price(block=block)
    
    @log(logger)
    @ttl_cache(ttl=36000)
    def get_price(self, token_in: AnyAddressType, block: Optional[Block] = None, protocol: Optional[str] = None) -> Optional[UsdPrice]:
        """
        Calculate a price based on Uniswap Router quote for selling one `token_in`.
        Always finds the deepest swap path for `token_in`.
        """
        token_in = convert.to_address(token_in)

        if protocol:
            return self.routers[protocol].get_price(token_in, block=block)

        for router in self.routers_by_depth(token_in, block=block):
            # tries each known router from most to least liquid
            # returns the first price we get back, almost always from the deepest router
            price = router.get_price(token_in, block=block)
            if price:
                return price
        
        if chain.id == Network.Mainnet:
            return self.get_price_v1(token_in, block)
        
        return None
    

    @log(logger)
    def deepest_router(self, token_in: AnyAddressType, block: Optional[Block] = None) -> Optional[UniswapRouterV2]:
        token_in = convert.to_address(token_in)

        for router in self.routers_by_depth(token_in, block=block):
            return router # will return first router in the dict, or None if no supported routers
        return None


    @log(logger)
    def routers_by_depth(self, token_in: str, block: Optional[Block] = None) -> Dict[UniswapRouterV2,str]:
        '''
        Returns a dict {router: pool} ordered by liquidity depth, greatest to least
        '''
        token_in = convert.to_address(token_in)

        pools_to_routers = {pool: router for router in self.routers.values() for pool in router.pools_for_token(token_in)}
        reserves = multicall_same_func_no_input(pools_to_routers, 'getReserves()((uint112,uint112,uint32))', block=block, return_None_on_failure=True)
        routers_by_depth = {}
        for router, pool, reserves in zip(pools_to_routers.values(), pools_to_routers.keys(), reserves):
            if reserves is None:
                continue
            if token_in == router.pools[pool]['token0']:
                routers_by_depth[reserves[0]] = {router: pool}
            elif token_in == router.pools[pool]['token1']:
                routers_by_depth[reserves[1]] = {router: pool}
        return {router: pool for balance in sorted(routers_by_depth, reverse=True) for router, pool in routers_by_depth[balance].items()}


uniswap = Uniswap()



import logging
from functools import cached_property
from typing import List, Optional, Union

from brownie import chain
from y.balancer.v1.v1 import BalancerV1
from y.balancer.v2.v2 import BalancerV2
from y.datatypes import UsdPrice
from y.decorators import log
from y.networks import Network
from y.typing import Block

logger = logging.getLogger(__name__)


class Balancer:
    def __init__(self) -> None:
        pass
    
    def __repr__(self) -> str:
        return "<Balancer>"
    
    @cached_property
    def versions(self) -> List[Union[BalancerV1, BalancerV2]]:
        return [v for v in [self.v1, self.v2] if v]

    @cached_property
    def v1(self) -> Optional[BalancerV1]:
        try: return BalancerV1()
        except ImportError: return None
    
    @cached_property
    def v2(self) -> Optional[BalancerV2]:
        try: return BalancerV2()
        except ImportError: return None

    @log(logger)
    def is_balancer_pool(self, token_address: str) -> bool:
        return any(v.is_pool(token_address) for v in self.versions)
    
    @log(logger)
    def get_pool_price(self, token_address: str, block: Optional[Block] = None) -> Optional[UsdPrice]:
        for v in self.versions:
            if v.is_pool(token_address):
                return UsdPrice(v.get_pool_price(token_address, block))

    @log(logger)
    def get_price(self, token_address: str, block: Optional[Block] = None) -> Optional[UsdPrice]:
        if self.is_balancer_pool(token_address):
            return self.get_pool_price(token_address, block=block)

        price = None    
        
        if ( # NOTE: Only query v2 if block queried > v2 deploy block plus some extra blocks to build up liquidity
            (chain.id == Network.Mainnet and (not block or block > 12272146 + 100000))
            or (chain.id == Network.Fantom and (not block or block > 16896080))
            ): 
            price = self.v2.get_token_price(token_address, block)
            if price:
                logger.debug(f"balancer v2 -> ${price}")
                return price

        if not price and chain.id == Network.Mainnet:      
            price = self.v1.get_token_price(token_address, block)
            if price:
                logger.debug(f"balancer v1 -> ${price}")
                return price
        

balancer = Balancer()



import logging
from functools import cached_property, lru_cache
from typing import Any, Optional, Set

from brownie import chain, convert
from multicall import Call
from y.classes.common import ERC20, ContractBase
from y.classes.singleton import Singleton
from y.constants import EEE_ADDRESS
from y.contracts import has_methods
from y.datatypes import UsdPrice
from y.decorators import log
from y.networks import Network
from y.typing import Block
from y.utils.logging import gh_issue_request
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)

TROLLERS = {
    Network.Mainnet: {
        "comp":             "0x3d9819210A31b4961b30EF54bE2aeD79B9c9Cd3B",
        "cream":            "0x3d5BC3c8d13dcB8bF317092d84783c2697AE9258",
        "ironbank":         "0xAB1c342C7bf5Ec5F02ADEA1c2270670bCa144CbB",
        "inverse":          "0x4dCf7407AE5C07f8681e1659f626E114A7667339",
        "unfederalreserve": "0x3105D328c66d8d55092358cF595d54608178E9B5",
    },
    Network.BinanceSmartChain: {
        "venus":            "0xfD36E2c2a6789Db23113685031d7F16329158384",
    },
    Network.Polygon: {
        "easyfi":           "0xcb3fA413B23b12E402Cfcd8FA120f983FB70d8E8",
        "apple":            "0x46220a07F071D1a821D68fA7C769BCcdA3C65430",
        "chumhum":          "0x1D43f6DA91e9EF6614dCe95bCef43E4d7b2bcFB5",
        "cream":            "0x20CA53E2395FA571798623F1cFBD11Fe2C114c24",
    },
    Network.Fantom: {
        "cream":            "0x4250A6D3BD57455d7C6821eECb6206F507576cD2",
        "scream":           "0x260E596DAbE3AFc463e75B6CC05d8c46aCAcFB09",
        "ola":              "0xD65eB596cFb5DE402678a12df651E0e588Dc3A81",
    },
    Network.Avalanche: {
        "vee":              "0xA67DFeD73025b0d61F2515c531dd8D25D4Cfd0Db",
        "vee2":             "0x43AAd7d8Bc661dfA70120865239529ED92Faa054",
        "vee3":             "0xeEf69Cab52480D2BD2D4A3f3E8F5CcfF2923f6eF",
        "cream":            "0x2eE80614Ccbc5e28654324a66A396458Fa5cD7Cc",
    },
    Network.Arbitrum: {
        "cream":            "0xbadaC56c9aca307079e8B8FC699987AAc89813ee",
        "neku":             "0xD5B649c7d27C13a2b80425daEe8Cb6023015Dc6B",
        "channels":         "0x3C13b172bf8BE5b873EB38553feC50F78c826284",
        "hund":             "0x0F390559F258eB8591C8e31Cf0905E97cf36ACE2",
    }
}.get(chain.id, {})


class CToken(ERC20):
    def __init__(self, address: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def get_price(self, block: Optional[Block] = None) -> UsdPrice:
        return UsdPrice(self.underlying_per_ctoken(block=block) * self.underlying.price(block=block))
    
    @cached_property
    @log(logger)
    def underlying(self) -> ERC20:
        underlying = self.has_method('underlying()(address)', return_response=True)

        # this will run for gas coin markets like cETH, crETH
        if not underlying: underlying = EEE_ADDRESS

        return ERC20(underlying)
    
    @log(logger)
    @lru_cache
    def underlying_per_ctoken(self, block: Optional[Block] = None) -> float:
        return self.exchange_rate(block=block) * 10 ** (self.decimals - self.underlying.decimals)
    
    @log(logger)
    @lru_cache
    def exchange_rate(self, block: Optional[Block] = None) -> float:
        method = 'exchangeRateCurrent()(uint)'
        try:
            exchange_rate = Call(self.address, [method], [[method,None]], block_id=block)()[method]
        except Exception as e:
            if 'borrow rate is absurdly high' in str(e):
                exchange_rate = 0
            else:
                raise
        return exchange_rate / 1e18
    

class Comptroller(ContractBase):
    def __init__(self, address: str = None, key: str = None) -> None:
        assert address or key,          'Must provide either an address or a key'
        assert not (address and key),   'Must provide either an address or a key, not both'

        if key: address = TROLLERS[key]
        else: key = [key for key in TROLLERS if address == TROLLERS[key]][0]

        self.address = convert.to_address(address)
        self.key = key
    
    def __repr__(self) -> str:
        return f"<Comptroller {self.key} '{self.address}'>"

    @log(logger)
    def __contains__(self, token_address: str) -> bool:
        return token_address in self.markets
    
    @cached_property
    def markets(self) -> Set[str]:
        response = self.has_method("getAllMarkets()(address[])", return_response=True)
        if not response:
            logger.error(f'had trouble loading markets for {self.__repr__()}')
            response = set()
        markets = {CToken(market) for market in response}
        logger.info(f"loaded {len(markets)} markets for {self.__repr__()}")
        return response


class Compound(metaclass = Singleton):
    def __init__(self) -> None:
        self.trollers = {
            protocol: Comptroller(troller)
            for protocol, troller
            in TROLLERS.items()
        }

    @log(logger)
    def is_compound_market(self, token_address: str) -> bool:
        if any(token_address in troller for troller in self.trollers.values()):
            return True

        # NOTE: Workaround for pools that have since been revoked
        result = has_methods(token_address, ['isCToken()(bool)','comptroller()(address)','underlying()(address)'])
        if result is True: self.__notify_if_unknown_comptroller(token_address)
        return result
    
    @log(logger)
    def get_price(self, token_address: str, block: Optional[Block] = None) -> UsdPrice:
        return CToken(token_address).get_price(block=block)

    @log(logger)
    def __contains__(self, token_address: str) -> bool:
        return self.is_compound_market(token_address)

    @log(logger)
    def __notify_if_unknown_comptroller(self, token_address: str) -> None:
        comptroller = raw_call(token_address,'comptroller()',output='address')
        if comptroller not in self.trollers.values():
            gh_issue_request(f'Comptroller {comptroller} is unknown to ypricemagic.', logger)


compound = Compound()



import logging
from typing import Optional

from brownie import chain
from y.balancer.v1.pool import BalancerV1Pool
from y.classes.singleton import Singleton
from y.constants import dai, usdc, wbtc, weth
from y.contracts import Contract, has_methods
from y.datatypes import UsdPrice
from y.decorators import log
from y.networks import Network
from y.prices import magic
from y.typing import Block
from y.utils.raw_calls import _decimals

EXCHANGE_PROXY = {
    Network.Mainnet: '0x3E66B66Fd1d0b02fDa6C811Da9E0547970DB2f21',
}.get(chain.id, None)

logger = logging.getLogger(__name__)

class BalancerV1(metaclass=Singleton):
    def __init__(self) -> None:
        self.exchange_proxy = Contract(EXCHANGE_PROXY) if EXCHANGE_PROXY else None
    
    def __str__(self) -> str:
        return "BalancerV1()"
    
    @log(logger)
    def is_pool(self, token_address) -> bool:
        return has_methods(token_address ,{"getCurrentTokens()(address[])", "getTotalDenormalizedWeight()(uint)", "totalSupply()(uint)"})
    
    @log(logger)
    def get_pool_price(self, token_address, block: Optional[Block] = None) -> UsdPrice:
        assert self.is_pool(token_address)
        return BalancerV1Pool(token_address).get_pool_price(block=block)

    @log(logger)
    def get_token_price(self, token_address, block: Optional[Block] = None) -> UsdPrice:
        out, totalOutput = self.get_some_output(token_address, block=block)
        if out: return (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block)
        # Can we get an output if we try smaller size?
        scale = 0.5
        out, totalOutput = self.get_some_output(token_address, block=block, scale=scale) 
        if out: return (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block) / scale
        # How about now? 
        scale = 0.1
        out, totalOutput = self.get_some_output(token_address, block=block, scale=scale)
        if out: return (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block) / scale
        else: return

    @log(logger)
    def check_liquidity_against(self, token_in, token_out, scale=1, block: Optional[Block] = None):
        output = self.exchange_proxy.viewSplitExactIn(
            token_in, token_out, 10 ** _decimals(token_in) * scale, 32 # NOTE: 32 is max
            , block_identifier = block
        )['totalOutput']
        return token_out, output

    @log(logger)
    def get_some_output(self, token_in, scale=1, block: Optional[Block] = None):
        try:
            out, totalOutput = self.check_liquidity_against(token_in, weth, block=block)
        except ValueError:
            try:
                out, totalOutput = self.check_liquidity_against(token_in, dai, block=block)
            except ValueError:
                try:
                    out, totalOutput = self.check_liquidity_against(token_in, usdc, block=block)
                except ValueError:
                    try:
                        out, totalOutput = self.check_liquidity_against(token_in, wbtc, block=block)
                    except ValueError:
                        out = None
                        totalOutput = None
        return out, totalOutput
